--[[――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    File: harness.lua
    Copyright 🄯 2021—2025 Van de Bugger.
    SPDX-License-Identifier: GPL-3.0-or-later
――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――]]--

function write( format, ... )
    print( string.format( format, ... ) )
end

-- Stringify value.
function _str( value )
    if type( value ) == 'string' then
        return '"' .. subst(
            value,
            '[%c"\\]',
            function ( chr )
                if chr == '\a' then
                    return '\\a'
                elseif chr == '\b' then
                    return '\\b'
                elseif chr == '\f' then
                    return '\\f'
                elseif chr == '\n' then
                    return '\\n'
                elseif chr == '\r' then
                    return '\\r'
                elseif chr == '\t' then
                    return '\\t'
                elseif chr == '\v' then
                    return '\\v'
                elseif chr == '\\' then
                    return '\\\\'
                elseif chr == '"' then
                    return '\\"'
                end
                local result = ''
                for i, k in ipairs( table.pack( string.byte( chr, 1, -1 ) ) ) do
                    result = result .. sprintf( '\\x%02X', k )
                end
                return result
            end
        ) .. '"'
    elseif type( value ) == 'table' then
        local keys = {}
        for k, v in pairs( value ) do
            keys[ # keys + 1 ] = k
        end
        table.sort( keys, function ( a, b )
            local ta = type( a )
            local tb = type( b )
            if ta == tb then
                if ta == 'boolean' then -- Lua can't compare boolean values
                    return ( a and 1 or 0 ) < ( b and 1 or 0 )
                end
                return a < b
            else
                return ta < tb
            end
        end )
        local items = {}
        for i, k in ipairs( keys ) do
            items[ # items + 1 ] = '[' .. _str( k ) .. ']=' .. _str( value[ k ] )
        end
        value = table.concat( items, ', ' )
        if # value == 0 then
            value = '{}'
        else
            value = '{ ' .. value .. ' }'
        end
    else
        value = tostring( value )
    end
    return value
end

function number_of_values( ... )
    local args = { ... }
    return # args
end

function mkdir( path )
    -- TODO: How to make intermidiate directories?
    local ok, msg, code = posix.mkdir( path )
    if not ok then
        if code == 17 and posix.stat( path, 'type' ) == 'directory' then
            -- It's ok: directory already exists.
        else
            -- msg includes both the problem path and the error message.
            error( string.format( "Can't create directory %s", msg ) )
        end
    end
end

ctx = {
    -- Open log file.
    open = function ( self, path )
        local topdir = rpm.expand( '%{_topdir}' )
        mkdir( topdir )
        self.path = topdir .. '/' .. path
        self.temp = self.path .. '.tmp'
        local err
        self.file, err = io.open( self.temp, 'w' )
        if err then
            error( string.format( "Can't open %s: %s", self.temp, err ) )
        end
        self.name   = ''
        self.total  = 0
        self.passed = 0
        self.failed = 0
        self.stack  = {}
    end,
    -- Wrtite a line to the log.
    _write = function ( self, format, ... )
        local str = string.format( format, ... )
        dbg( '%s', str )
        self.file:write( str .. '\n' )
    end,
    -- Close the log file.
    close = function ( self )
        self.file:close()
        local ok, err = os.rename( self.temp, self.path )
        if not ok then
            error( string.format( "Can't rename %s to %s: %s", self.temp, self.path, err ) )
        end
    end,
    -- Write test line: test status, test number, and description.
    test = function ( self, ok, description )
        self.total = self.total + 1
        if ok then
            self.passed = self.passed + 1
        else
            self.failed = self.failed + 1
        end
        self:_write(
            '%s %d%s',
            ok and 'ok' or 'not ok',
            self.total,
            description and ' - ' .. description or ''
        )
        return ok
    end,
    -- Wite diagnostic message.
    diag = function ( self, format, ... )
        self:_write( '# %s', string.format( format, ... ) )
        return true
    end,
    -- Write the test plan.
    plan = function( self )
        self:_write( '1..%d', self.total )
    end,
    push = function ( self, name )
        local current = {
            name   = self.name,
            passed = self.passed,
            failed = self.failed,
        }
        self.stack[ # self.stack + 1 ] = current
        self.name = ( # self.name > 0 and self.name .. '/' or '' ) .. name
        self.passed = 0
        self.failed = 0
    end,
    pop = function ( self )
        local current = {
            name   = self.name,
            passed = self.passed,
            failed = self.failed,
        }
        local top = self.stack[ # self.stack ]
        self.name   = top.name
        self.passed = self.passed + top.passed
        self.failed = self.failed + top.failed
        self.stack[ # self.stack ] = nil
        return current
    end,
}

function diag( format, ... )
    return ctx:diag( format, ... )
end

function fail( description )
    return not ctx:test( false, description )
end

function ok( condition, description )
    return ctx:test( condition, description )
end

function eq( actual, expected, description )
    local ta = type( actual )
    local te = type( expected )
    local sa = _str( actual )
    local se = _str( expected )
    return ok( ta == te and sa == se, description )
        or diag( 'act: %s', sa ) and diag( 'exp: %s', se ) and false
end

function start( path )
    ctx:open( path )
end

function test( name, body )
    ctx:push( name )
    body()
    local current = ctx:pop()
    return ok( current.failed == 0, current.name )
end

function done()
    ctx:plan()
    ctx:diag( '%d tests run', ctx.total )
    ctx:diag( '%d tests passed', ctx.passed )
    ctx:diag( '%d tests failed', ctx.failed )
    ctx:close()
    set_env( 'tests_total',  ctx.total  )
    set_env( 'tests_passed', ctx.passed )
    set_env( 'tests_failed', ctx.failed )
end

-- end of file --
