--[[――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    File: tests.lua
    Copyright 🄯 2021—2025 Van de Bugger.
    SPDX-License-Identifier: GPL-3.0-or-later
――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――]]--

if not get_env( 'tests_total' ) then

--[[--
    rpmbuild evaluates the file multiple times, so I added this guard. Lua variable is not suitable
    since the Lua interpreter is reset and all the Lua variables are reset too, but environment
    variable seems to work.
--]]--

start( 'tests.log' )

test( 'string manupulation routines', function ()
----------------------------------------------------------------------------------------------------

    test( 'subst', function ()
        eq( subst( 'abcdefabcdef', 'b', 'x'           ), 'axcdefaxcdef', 'one subst' )
        eq( subst( 'abcdefabcdef', 'b', 'x', 'd', 'y' ), 'axcyefaxcyef', 'two subst' )
        eq( subst( 'abcdefabcdef', 'b', 'x', 'd'      ), 'axcefaxcef',   'odd number of args' )
        eq( subst( 'abcdefabcdef'                     ), 'abcdefabcdef', 'no substs' )
        eq( number_of_values( subst( 'abcdefabcdef', 'a', 'z' ) ), 1,    'one returned value' )
    end )

    test( 'escape', function ()
        eq( escape( 'abc'    ), 'abc',       'no %'   )
        eq( escape( '%abc'   ), '%%abc',     'one %'  )
        eq( escape( '%a%b%c' ), '%%a%%b%%c', 'many %' )
    end )

    test( 'quote', function ()
        eq( quote( "abc"   ),  "'abc'" )
        eq( quote( "a'bc"  ), "'a'\\''bc'" )
        eq( quote( "a'b'c" ), "'a'\\''b'\\''c'" )
    end )

    test( 'trim', function ()
        eq( trim( 'abc'   ),  'abc' )
        eq( trim( ' abc'  ),  'abc' )
        eq( trim( 'abc '  ),  'abc' )
        eq( trim( ' abc ' ),  'abc' )
        eq( trim( '\n\tabc\t\n' ), 'abc' )
        eq( trim( '\n\ta\n\nb\t\tc\t\n' ), 'a\n\nb\t\tc' )
    end )

    test( 'chomp', function ()
        eq( chomp( ''      ),  ''    )
        eq( chomp( '\n'    ),  ''    )
        eq( chomp( '\n\n'  ),  '\n'  )
        eq( chomp( '\n '    ), '\n ' )
    end )

    test( 'shrink', function ()
        eq( shrink( '        abc' ),  'abc',   'leading spaces'    )
        eq( shrink( '\t\t\t\tabc' ),  'abc',   'leading tabs'      )
        eq( shrink( '\n\n\n\nabc' ),  'abc',   'leading newlines'  )
        eq( shrink( 'abc        ' ),  'abc',   'trailing spaces'   )
        eq( shrink( 'abc\t\t\t\t' ),  'abc',   'trailing tabs'     )
        eq( shrink( 'abc\n\n\n\n' ),  'abc',   'trailing newlines' )
        eq( shrink( 'a    b    c' ),  'a b c', 'inner spaces'      )
        eq( shrink( 'a\t\tb\t\tc' ),  'a b c', 'inner tabs'        )
        eq( shrink( 'a\n\nb\n\nc' ),  'a b c', 'inner newlines'    )
    end )

    test( 'join', function ()
        test( 'arrays', function ()
            eq( join( ',', {} ), '',                          'empty array'    )
            eq( join( ',', { 'a', 'b', 'c' } ), 'a,b,c',      'string values'  )
            eq( join( ',', { 1, 3, 10 }      ), '1,3,10',     'numeric values' )
            eq( join( ',', { true, false }   ), 'true,false', 'boolean values' )
            eq( join( ',', { 'a', true, 10 } ), 'a,true,10',  'mixed values'   )
        end )
        test( 'hashes', function ()
            eq( join( ',', { a = 'a', z = 'b', h = 'c' } ),         'a,c,b', 'string keys'  )
            eq( join( ',', { [ true ] = 'a', [ false ] = 'b' } ),   'b,a',   'boolean keys' )
            eq( join( ',', { 'a', [ false ] = 'b', [ 'z' ] = 8 } ), 'b,a,8', 'mixed keys'   )
        end )
    end )

    test( 'split', function ()
        test( 'sep is colon', function ()
            local s = ':ab:cd::ef::'
            eq( split( ':', s, 1  ), { ':ab:cd::ef::'                   }, 'limit 1'  )
            eq( split( ':', s, 2  ), { '', 'ab:cd::ef::'                }, 'limit 2'  )
            eq( split( ':', s, 3  ), { '', 'ab', 'cd::ef::'             }, 'limit 3'  )
            eq( split( ':', s, 4  ), { '', 'ab', 'cd', ':ef::'          }, 'limit 4'  )
            eq( split( ':', s, 5  ), { '', 'ab', 'cd', '', 'ef::'       }, 'limit 5'  )
            eq( split( ':', s, 6  ), { '', 'ab', 'cd', '', 'ef', ':'    }, 'limit 6'  )
            eq( split( ':', s, 7  ), { '', 'ab', 'cd', '', 'ef', '', '' }, 'limit 7'  )
            eq( split( ':', s, 8  ), { '', 'ab', 'cd', '', 'ef', '', '' }, 'limit 8'  )
            eq( split( ':', s, 0  ), { '', 'ab', 'cd', '', 'ef'         }, 'limit 0'  )
            eq( split( ':', s     ), { '', 'ab', 'cd', '', 'ef'         }, 'no limit' )
            eq( split( ':', s, -1 ), { '', 'ab', 'cd', '', 'ef', '', '' }, 'limit -1' )
        end )
        test( 'sep is single space', function ()
            -- single space sep is a special case: leading whitespace stripped, sep is %s+:
            local s = '\nab cd\t ef\n\n'
            eq( split( ' ', s, 1  ), { 'ab cd\t ef\n\n'                 }, 'limit 1'  )
            eq( split( ' ', s, 2  ), { 'ab', 'cd\t ef\n\n'              }, 'limit 2'  )
            eq( split( ' ', s, 3  ), { 'ab', 'cd', 'ef\n\n'             }, 'limit 3'  )
            eq( split( ' ', s, 4  ), { 'ab', 'cd', 'ef', ''             }, 'limit 4'  )
            eq( split( ' ', s, 5  ), { 'ab', 'cd', 'ef', ''             }, 'limit 5'  )
            eq( split( ' ', s, 0  ), { 'ab', 'cd', 'ef'                 }, 'limit 0'  )
            eq( split( ' ', s     ), { 'ab', 'cd', 'ef'                 }, 'no limit' )
            eq( split( ' ', s, -1 ), { 'ab', 'cd', 'ef', ''             }, 'limit -1' )
        end )
        test( 'str is empty', function ()
            local s = ''
            eq( split( ':', s, -1 ), {}, 'limit -1' )
            eq( split( ':', s, 0  ), {}, 'limit 0'  )
            eq( split( ':', s, 1  ), {}, 'limit 1'  )
            eq( split( ':', s, 2  ), {}, 'limit 2'  )
            eq( split( ':', s     ), {}, 'no limit' )
        end )
        test( 'sep is empty', function ()
            local s = 'abc'
            eq( split( '', s, 1  ), { 'abc'                            }, 'limit 1'  )
            eq( split( '', s, 2  ), { 'a', 'bc'                        }, 'limit 2'  )
            eq( split( '', s, 3  ), { 'a', 'b', 'c'                    }, 'limit 3'  )
            eq( split( '', s, 4  ), { 'a', 'b', 'c', ''                }, 'limit 4'  )
            eq( split( '', s, 5  ), { 'a', 'b', 'c', ''                }, 'limit 5'  )
            eq( split( '', s, 0  ), { 'a', 'b', 'c'                    }, 'limit 0'  )
            eq( split( '', s     ), { 'a', 'b', 'c'                    }, 'no limit' )
            eq( split( '', s, -1 ), { 'a', 'b', 'c', ''                }, 'limit -1' )
        end )
    end )

    test( 'uc', function ()
        eq( uc( 'abc' ), 'ABC' )
    end )

    test( 'lc', function ()
        eq( lc( 'ABC' ), 'abc' )
    end )

    test( 'tc', function ()
        eq( tc( 'ABC' ), 'Abc' )
        eq( tc( 'ABC XYZ' ), 'Abc Xyz' )
        eq( tc( ' ABC\tXYZ\nijk' ), ' Abc\tXyz\nIjk' )
    end )

    test( 'sprintf', function ()
        eq( sprintf( '%03d %s', 10, 'abc' ), '010 abc' )
    end )

    test( 'ident', function ()
        eq( ident( 'Foo Bar'    ), 'foo-bar' )
        eq( ident( 'Foo', 'Bar' ), 'foo-bar' )
    end )

    test( 'prepend', function ()
        eq( prepend( 'xxx', ''            ), ''                     )
        eq( prepend( 'xxx', 'abc'         ), 'xxxabc'               )
        eq( prepend( 'xxx', 'abc\n'       ), 'xxxabc\n'             )
        eq( prepend( 'xxx', 'abc\n '      ), 'xxxabc\nxxx '         )
        eq( prepend( 'xxx', 'abc\nijk'    ), 'xxxabc\nxxxijk'       )
        eq( prepend( 'xxx', 'abc\nijk\n'  ), 'xxxabc\nxxxijk\n'     )
        eq( prepend( 'xxx', 'abc\nijk\n ' ), 'xxxabc\nxxxijk\nxxx ' )
    end )

    test( 'reflow', function ()
        local f = reflow
        eq( f( 'short text', 10 ), 'short text', 'one line: no wrap' )
        eq( f( 'short text', 9  ), 'short\ntext', 'one line: wrap' )
        eq( f( 'this is short text', 7  ), 'this is\nshort\ntext', 'one line: multiwrap' )
        eq( f( '\n\n\t\t  leading whitespace dropped', 20 ), 'leading whitespace\ndropped' )
        eq( f( 'trailing whitespace dropped \t\n \t\n', 20 ), 'trailing whitespace\ndropped' )
        eq( f( 'inner \t\t\t whitespace  \t\n shrinked', 20 ), 'inner whitespace\nshrinked' )
        eq( f( 'empty lines\n\n\n\n\n\n\nshrinked too', 10 ), 'empty\nlines\n\nshrinked\ntoo' )
        eq( f( 'just fit fit just just fit', 8 ), 'just fit\nfit just\njust fit' )
        eq( f( 'tooooooo   looooong   woooooords', 5 ), 'tooooooo\nlooooong\nwoooooords' )
    end )

    test( 'str( nil )', function ()
        eq( str(),            'nil'       )
        eq( str( nil ),       'nil'       )
    end )
    test( 'str( boolean )', function ()
        eq( str( false ),     'false'     )
        eq( str( true  ),     'true'      )
    end )
    test( 'str( number )', function ()
        eq( str( 0       ),   '0'         )
        eq( str( 1       ),   '1'         )
        eq( str( 1000000 ),   '1000000'   )
        eq( str( 1.0       ), '1.0'       )
        eq( str( -0.999999 ), '-0.999999' )
    end )
    test( 'str( string )', function ()
        eq( str( ''      ),  "''"        )
        eq( str( 'abc'   ),  "'abc'"     )
        eq( str( '"'     ),  "'\"'"      )
        eq( str( '\t'    ),  "'\\t'"     )
        eq( str( '\n'    ),  "'\\n'"     )
        eq( str( '\''    ),  "'\\''"     )
    end )
    test( 'str( table ) # array', function ()
        eq( str( {} ),                "{}"                                             )
        eq( str( { false, true } ),   "{\n    1 → false,\n    2 → true,\n}"            )
        eq( str( { false, 0, "x" } ), "{\n    1 → false,\n    2 → 0,\n    3 → 'x',\n}" )
    end )
    test( 'str( table ) # hash', function ()
        eq( str( { [ 'x' ] = false, [ 1 ] = 'z' } ), "{\n    1 → 'z',\n    'x' → false,\n}" )
        eq( str( { [ '1' ] = 'y',   [ 1 ] = 'z' } ), "{\n    1 → 'z',\n    '1' → 'y',\n}"   )
    end )
    test( 'str( function ) ', function ()
        local s = str( function () end )
        return string.match( s, '^function: 0x[0-9a-f]+$' ) or fail() and diag( 'str: %s', s )
    end )
end )

test( 'Array manipulation routines', function ()
----------------------------------------------------------------------------------------------------

    test( 'push', function ()
        local a = {}
        push( a, 'a' )
        eq( a, { 'a' } )
        push( a, 'b' )
        eq( a, { 'a', 'b' } )
        push( a, 'c' )
        eq( a, { 'a', 'b', 'c' } )
        push( a, 'd', 'e', 'f' )
        eq( a, { 'a', 'b', 'c', 'd', 'e', 'f' } )
    end )

    test( 'pop', function ()
        local a = { 'a', 'b', 'c' }
        eq( pop( a ), 'c' )
        eq( a, { 'a', 'b' } )
        eq( pop( a ), 'b' )
        eq( a, { 'a' } )
        eq( pop( a ), 'a' )
        eq( a, {} )
        eq( pop( a ), nil, 'empty array' )
        eq( a, {} )
    end )

    test( 'shift', function ()
        local a = { 'a', 'b', 'c' }
        eq( shift( a ), 'a' )
        eq( a, { 'b', 'c' } )
        eq( shift( a ), 'b' )
        eq( a, { 'c' } )
        eq( shift( a ), 'c' )
        eq( a, {} )
        eq( shift( a ), nil, 'empty array' )
        eq( a, {} )
    end )

    test( '_sort', function ()
        test( 'boolean', function ()
            local a = { true, false }
            eq( _sort( a ), { false, true } )
            eq( a, { false, true } )
        end )
        test( 'numeric', function ()
            local a = { 10, 9, 1 }
            eq( _sort( a ), { 1, 9, 10 } )
            eq( a, { 1, 9, 10 } )
        end )
        test( 'string', function ()
            local a = { 'c', 'b', 'a' }
            eq( _sort( a ), { 'a', 'b', 'c' } )
            eq( a, { 'a', 'b', 'c' } )
            a = { '10', '9', '1' }
            eq( _sort( a ), { '1', '10', '9' } )
            eq( a, { '1', '10', '9' } )
        end )
        test( 'mixed', function ()
            local a = { 'z', 'a', '10', 100, 10, 1, false, true }
            eq( _sort( a ), { false, true, 1, 10, 100, '10', 'a', 'z' } )
            eq( a, { false, true, 1, 10, 100, '10', 'a', 'z' } )
        end )
    end )

end )

test( 'Hash manipulation routines', function ()

    test( 'hash', function ()
        eq( hash( { 1, 2, 3 }          ), { [ 1 ] = false, [ 2 ] = false, [ 3 ] = false } )
        eq( hash( { 1, 2, 3 }, true    ), { [ 1 ] = true,  [ 2 ] = true,  [ 3 ] = true  } )
        eq( hash( { 1, 2, 3 }, 1.5     ), { [ 1 ] = 1.5,   [ 2 ] = 1.5,   [ 3 ] = 1.5   } )
        eq( hash( { 1, 2, 3 }, 'x'     ), { [ 1 ] = 'x',   [ 2 ] = 'x',   [ 3 ] = 'x'   } )
        eq( hash( { 1, 2, 7 }, 'z'     ), { [ 1 ] = 'z',   [ 2 ] = 'z',   [ 7 ] = 'z'   } )
        eq( hash( { 1, 7, 'a' }, 1     ), { [ 1 ] = 1,     [ 7 ] = 1,     [ 'a' ] = 1   } )
        eq( hash( { false, 1, 'a' }, 1 ), { [ false ] = 1, [ 1 ] = 1,     [ 'a' ] = 1   } )
    end )

    test( 'keys & vals & sort', function ()
        test( 'boolean', function ()
            local h = { [ true ] = '1', [ false ] = '0' }
            eq( keys( h ), { false, true } )
            eq( vals( h ), { '0', '1' } )
            eq( sort( h ), { '0', '1' } )
            eq( h, { [ true ] = '1', [ false ] = '0' } )
        end )
        test( 'boolean', function ()
            local h = { [ 1 ] = 'a', [ 2 ] = 'b', [ 10 ] = 'j' }
            eq( keys( h ), { 1, 2, 10 } )
            eq( vals( h ), { 'a', 'b', 'j' } )
            eq( sort( h ), { 'a', 'b', 'j' } )
            eq( h, { [ 1 ] = 'a', [ 2 ] = 'b', [ 10 ] = 'j' } )
        end )
        test( 'boolean', function ()
            local h = { a = 1, b = 2, j = 10 }
            eq( keys( h ), { 'a', 'b', 'j' } )
            eq( vals( h ), { 1, 2, 10 } )
            eq( sort( h ), { 1, 2, 10 } )
            eq( h, { a = 1, b = 2, j = 10 } )
        end )
        test( 'boolean', function ()
            local h = { [ false ] = 0, [ true ] = 1, [ 1 ] = 'a', [ '1' ] = true, [ 'b' ] = 2 }
            eq( keys( h ), { false, true, 1,  '1',  'b'  } )
            eq( vals( h ), { 0,     1,    'a', true, 2   } )
            eq( sort( h ), { true,  0,     1,    2,  'a' } )
            eq( h, { [ false ] = 0, [ true ] = 1, [ 1 ] = 'a', [ '1' ] = true, [ 'b' ] = 2 } )
        end )
    end )

    test( 'map', function ()
        test( 'value', function ()
            local h = { 1, 2, 5 }
            eq( map( function ( v ) return v + 1 end, h ), { 2, 3, 6 } )
            eq( h, { 1, 2, 5 } )
        end )
        test( 'value & key', function ()
            local h = { 1, 2, 5 }
            eq( map( function ( v, k ) return v + k end, h ), { 2, 4, 8 } )
            eq( h, { 1, 2, 5 } )
        end )
        test( 'mixed keys', function ()
            local h = { [ 'a' ] = 1, [ 1 ] = 2, [ true ] = 5 }
            eq( map( function ( v ) return v + 2 end, h ), { 7, 4, 3 } )
            eq( h, { [ 'a' ] = 1, [ 1 ] = 2, [ true ] = 5 } )
        end )
    end )

    test( 'grep', function ()
        test( 'value argument', function ()
            local h = { 5, 6, 1, 2 }
            eq( grep( function ( v ) return v % 2 == 0 end, h ), { 6, 2 } )
            eq( h, { 5, 6, 1, 2 } )
        end )
        test( 'value & key arguments', function ()
            local h = { 5, 6, 1, 2 }
            eq( grep( function ( v, k ) return k < 3 end, h ), { 5, 6 } )
            eq( h, { 5, 6, 1, 2 } )
        end )
        test( 'mixed keys', function ()
            local h = { [ 'a' ] = 5, [ 2 ] = 6, [ true ] = 1, [ '2' ] = 2 }
            eq( grep( function ( v ) return v % 2 == 1 end, h ), { 1, 5 } )
            eq( h, { [ 'a' ] = 5, [ 2 ] = 6, [ true ] = 1, [ '2' ] = 2 } )
        end )
        -- TODO: value, not function
    end )

end )

test( 'RPM primitive routines', function ()
----------------------------------------------------------------------------------------------------

    test( 'expand', function ()
        eq( expand( ''          ), '',               'empty string'     )
        eq( expand( 'abc'       ), 'abc',            'non-empty string' )
        eq( expand( '%{nil}'    ), '',               'empty macro '     )
        eq( expand( '%{name}'   ), 'vdb-rpm-macros', 'non-empty macro'  )
        eq( expand( '%{foobar}' ), '%{foobar}',      'undefined macro'  )
    end )

    test( 'defined', function ()
        eq( defined( 'nil'       ), true  )
        eq( defined( 'name'      ), true  )
        eq( defined( 'undefined_macro' ), false )
    end )

    test( 'set', function ()
        test( 'types', function ()
            set( 'foo', 'bar' )
            eq( rpm.expand( '%{foo}' ), 'bar' )
            set( 'foo', 'gaga' )
            eq( rpm.expand( '%{foo}' ), 'gaga' )
            set( 'foo' )
            eq( rpm.expand( '%{foo}' ), '', 'nil converted to empty string' )
            set( 'foo', false )
            eq( rpm.expand( '%{foo}' ), '0', 'false converted to 0' )
            set( 'foo', true )
            eq( rpm.expand( '%{foo}' ), '1', 'true converted to 0' )
        end )
        test( 'values', function ()
            set( 'bar', 'line one\n\nline two' )
            eq( rpm.expand( '%{bar}' ), 'line one\n\nline two', 'inner newlines are preserved' )
            set( 'bar', '\n\n\n1\n\n2' )
            eq( rpm.expand( '%{bar}' ), '\n\n\n1\n\n2', 'leading newlines are preserved too' )
            set( 'bar', '1\n\n2\n\n\n' )
            eq( rpm.expand( '%{bar}' ), '1\n\n2', 'but trailing newlines are not' )
        end )
    end )

    test( 'unset', function ()
        eq( rpm.expand( '%{foo}' ), '1' )
        unset( 'foo' )
        eq( rpm.expand( '%{foo}' ), '0' )
        unset( 'foo' )
        eq( rpm.expand( '%{foo}' ), '' )
        unset( 'foo' )
        eq( rpm.expand( '%{foo}' ), 'gaga' )
        unset( 'foo' )
        eq( rpm.expand( '%{foo}' ), 'bar' )
        unset( 'foo' )
        eq( rpm.expand( '%{foo}' ), '%{foo}' )
    end )

    test( 'get', function ()

        -- Undefined macro.
        eq( get( 'foo'        ),       nil, 'undef, nil'       )
        eq( get( 'foo',    '' ),        '', 'undef, empty'     )
        eq( get( 'foo',   'x' ),       'x', 'undef, non-empty' )
        eq( get( 'foo',     0 ),         0, 'undef, 0'         )
        eq( get( 'foo',     1 ),         1, 'undef, 1'         )
        eq( get( 'foo', false ),     false, 'undef, false'     )
        eq( get( 'foo',  true ),      true, 'undef, true'      )

        -- Empty macro.
        rpm.define( 'foo %{nil}' )
        eq( get( 'foo',    '' ),        '', 'empty, empty'     )
        eq( get( 'foo',   'x' ),        '', 'empty, x'         )
        eq( get( 'foo',    0  ),         0, 'empty, 0'         )
        eq( get( 'foo',    1  ),         1, 'empty, 1'         )
        eq( get( 'foo', false ),     false, 'empty, false'     )
        eq( get( 'foo',  true ),     false, 'empty, true'      )

        -- Zero.
        rpm.define( 'foo 0' )
        eq( get( 'foo'        ),       '0', 'zero, nil'       )
        eq( get( 'foo',    '' ),       '0', 'zero, empty'     )
        eq( get( 'foo',   'x' ),       '0', 'zero, x'         )
        eq( get( 'foo',    0  ),         0, 'zero, 0'         )
        eq( get( 'foo',    1  ),         0, 'zero, 1'         )
        eq( get( 'foo', false ),     false, 'zero, false'     )
        eq( get( 'foo',  true ),     false, 'zero, true'      )

        -- One.
        rpm.define( 'foo 1' )
        eq( get( 'foo'        ),       '1', 'one, nil'        )
        eq( get( 'foo',    '' ),       '1', 'one, empty'      )
        eq( get( 'foo',   'x' ),       '1', 'one, x'          )
        eq( get( 'foo',    0  ),         1, 'one, 0'          )
        eq( get( 'foo',    1  ),         1, 'one, 1'          )
        eq( get( 'foo', false ),      true, 'one, false'      )
        eq( get( 'foo',  true ),      true, 'one, true'       )

        -- Ten.
        rpm.define( 'foo 10' )
        eq( get( 'foo'        ),      '10', 'ten, nil'        )
        eq( get( 'foo',    '' ),      '10', 'ten, empty'      )
        eq( get( 'foo',   'x' ),      '10', 'ten, x'          )
        eq( get( 'foo',    0  ),        10, 'ten, 0'          )
        eq( get( 'foo',    1  ),        10, 'ten, 1'          )
        eq( get( 'foo', false ),      true, 'ten, false'      )
        eq( get( 'foo',  true ),      true, 'ten, true'       )

        -- NAN.
        rpm.define( 'foo 10a' )
        eq( get( 'foo'        ),     '10a', 'NAN, nil'        )
        eq( get( 'foo',    '' ),     '10a', 'NAN, empty'      )
        eq( get( 'foo',   'x' ),     '10a', 'NAN, x'          )
        eq( get( 'foo',    0  ),         0, 'NAN, 0'          )
        eq( get( 'foo',    1  ),         1, 'NAN, 1'          )
        eq( get( 'foo', false ),      true, 'NAN, false'      )
        eq( get( 'foo',  true ),      true, 'NAN, true'       )

    end )

    -- TODO: with
    -- TODO: spec

end )

test( 'OS and filesystem routines', function ()
----------------------------------------------------------------------------------------------------

    test( 'get_env', function ()

        -- Unset var
        eq( get_env( 'unset-var'         ),   nil, 'unset, nil'   )
        eq( get_env( 'unset-var',     '' ),    '', 'unset, empty' )
        eq( get_env( 'unset-var',      0 ),     0, 'unset, 0'     )
        eq( get_env( 'unset-var',      1 ),     1, 'unset, 1'     )
        eq( get_env( 'unset-var',  false ), false, 'unset, false' )
        eq( get_env( 'unset-var',   true ),  true, 'unset, true'  )

        -- Empty var
        posix.setenv( 'VDB_TEST', '', true )
        eq( get_env( 'VDB_TEST'         ),    '', 'empty, nil'    )
        eq( get_env( 'VDB_TEST',    'x' ),    '', 'empty, x'      )
        eq( get_env( 'VDB_TEST',      0 ),     0, 'empty, 0'      )
        eq( get_env( 'VDB_TEST',      1 ),     1, 'empty, 1'      )
        eq( get_env( 'VDB_TEST',  false ), false, 'empty, false'  )
        eq( get_env( 'VDB_TEST',   true ), false, 'empty, true'   )

        -- Zero
        posix.setenv( 'VDB_TEST', '0', true )
        eq( get_env( 'VDB_TEST'         ),   '0', 'zero, nil'     )
        eq( get_env( 'VDB_TEST',    'x' ),   '0', 'zero, x'       )
        eq( get_env( 'VDB_TEST',      0 ),     0, 'zero, 0'       )
        eq( get_env( 'VDB_TEST',      1 ),     0, 'zero, 1'       )
        eq( get_env( 'VDB_TEST',  false ), false, 'zero, false'   )
        eq( get_env( 'VDB_TEST',   true ), false, 'zero, true'    )

        -- One
        posix.setenv( 'VDB_TEST', '1', true )
        eq( get_env( 'VDB_TEST'         ),   '1', 'one, nil'      )
        eq( get_env( 'VDB_TEST',    'x' ),   '1', 'one, x'        )
        eq( get_env( 'VDB_TEST',      0 ),     1, 'one, 0'        )
        eq( get_env( 'VDB_TEST',      1 ),     1, 'one, 1'        )
        eq( get_env( 'VDB_TEST',  false ),  true, 'one, false'    )
        eq( get_env( 'VDB_TEST',   true ),  true, 'one, true'     )

        -- Ten
        posix.setenv( 'VDB_TEST', '10', true )
        eq( get_env( 'VDB_TEST'         ),  '10', 'ten, nil'      )
        eq( get_env( 'VDB_TEST',    'x' ),  '10', 'ten, x'        )
        eq( get_env( 'VDB_TEST',      0 ),    10, 'ten, 0'        )
        eq( get_env( 'VDB_TEST',      1 ),    10, 'ten, 1'        )
        eq( get_env( 'VDB_TEST',  false ),  true, 'ten, false'    )
        eq( get_env( 'VDB_TEST',   true ),  true, 'ten, true'     )

        -- NAN
        posix.setenv( 'VDB_TEST', '10a', true )
        eq( get_env( 'VDB_TEST'         ), '10a', 'NAN, nil'      )
        eq( get_env( 'VDB_TEST',    'x' ), '10a', 'NAN, x'        )
        eq( get_env( 'VDB_TEST',      0 ),     0, 'NAN, 0'        )
        eq( get_env( 'VDB_TEST',      1 ),     1, 'NAN, 1'        )
        eq( get_env( 'VDB_TEST',  false ),  true, 'NAN, false'    )
        eq( get_env( 'VDB_TEST',   true ),  true, 'NAN, true'     )

    end )

    test( 'set_env', function ()

        -- Unset by default.
        posix.setenv( 'VDB_TEST', '1', true )
        set_env( 'VDB_TEST' )
        eq( get_env( 'VDB_TEST' ), nil )

        -- Unset with nil.
        posix.setenv( 'VDB_TEST', '1', true )
        set_env( 'VDB_TEST', nil )
        eq( get_env( 'VDB_TEST' ), nil )

        -- Set, overwrite by default.
        posix.setenv( 'VDB_TEST', '1', true )
        set_env( 'VDB_TEST', 2 )
        eq( get_env( 'VDB_TEST' ), '2' )

        -- Set, overwrite by default, not existing.
        posix.unsetenv( 'VDB_TEST' )
        set_env( 'VDB_TEST', 2 )
        eq( get_env( 'VDB_TEST' ), '2' )

        -- Set, overwrite with true.
        posix.setenv( 'VDB_TEST', '1', true )
        set_env( 'VDB_TEST', 2, true )
        eq( get_env( 'VDB_TEST' ), '2' )

        -- Set, overwrite with true, not existing.
        posix.unsetenv( 'VDB_TEST' )
        set_env( 'VDB_TEST', 2, true )
        eq( get_env( 'VDB_TEST' ), '2' )

        -- Set, no overwrite.
        posix.setenv( 'VDB_TEST', '1', true )
        set_env( 'VDB_TEST', 2, false )
        eq( get_env( 'VDB_TEST' ), '1' )

        -- Set, no overwrite, not existing.
        posix.unsetenv( 'VDB_TEST' )
        set_env( 'VDB_TEST', 2, false )
        eq( get_env( 'VDB_TEST' ), '2' )

    end )

    test( 'exist', function ()
        eq( exist( '/' ),                    true,  'root directory'         )
        eq( exist( '.' ),                    true,  'current directory'      )
        eq( exist( '..' ),                   true,  'parent directory'       )
        eq( exist( '/non-existent-path' ),   false, 'non-existent directory' )
        eq( exist( rpm.expand( '%{S:0}' ) ), true,  'existing file'          )
        eq( exist( rpm.expand( '%{S:1}' ) ), true,  'existing file'          )
    end )

    test( 'which', function ()
        eq( which( 'sh' ),                          '/usr/bin/sh',  'dirs omitted'      )
        eq( which( 'sh', { '/bin', '/usr/bin' } ),  '/bin/sh',      'dirs specified'    )
    end )

    -- TODO: rename

    test( 'slurp', function ()
        local dir  = rpm.expand( '%{_builddir}' )
        mkdir( dir )
        local path = dir .. '/test.txt'
        local file, err = io.open( path, 'w' )
        if err then
            error( string.format( "Can't open %s: %s", path, err ) )
        end
        file:write( 'one\ntwo\nthree\n\n\n' )
        file:close()
        eq( slurp( path, 1 ), { 'one'                         }, 'one line' )
        eq( slurp( path, 2 ), { 'one', 'two'                  }, 'two lines' )
        eq( slurp( path, 3 ), { 'one', 'two', 'three'         }, 'three lines' )
        eq( slurp( path, 4 ), { 'one', 'two', 'three', ''     }, 'four lines' )
        eq( slurp( path, 5 ), { 'one', 'two', 'three', '', '' }, 'five lines' )
        eq( slurp( path, 5 ), { 'one', 'two', 'three', '', '' }, 'all five lines' )
        eq( slurp( path ), { 'one', 'two', 'three' },
            'entire file read, but trailing empty lines are dropped'
        )
        eq( slurp( path, 0 ), { 'one', 'two', 'three' },
            'entire file read, but trailing empty lines are dropped'
        )
        eq( slurp( path, -1 ), { 'one', 'two', 'three', '', '' },
            'entire file read, but trailing empty lines are not dropped'
        )
    end )

    test( 'spurt', function ()
        local path = rpm.expand( '%{_builddir}/spurt.txt' )
        spurt( path, 'line 1', 'line 2' )
        eq( exist( path ), true,  'make sure the file exists' )
        eq( slurp( path ), { 'line 1', 'line 2' }, 'check contents' )
        spurt( path, '1', '2', '3' )
        eq( exist( path ), true,  'make sure the file exists again' )
        eq( slurp( path ), { '1', '2', '3' }, 'check contents again' )
        spurt( path )
        eq( exist( path ), true,  'make sure the file exists one more time' )
        eq( slurp( path ), {}, 'check contents one more time' )
    end )

    -- TODO: shell

end )

test( 'External programs', function ()
----------------------------------------------------------------------------------------------------

    test( 'find_rpm', function()
        local v = find_rpm()
        ok( string.match( v, '^/' ) ) -- v is absolute path
        ok( string.match( v, '/bin/rpm$' ) ) -- v ends with '/bin/rpm'
    end )

    test( 'rpm', function ()
        -- I do not know which `install` version is available in the system,
        -- so check is rather weak.
        local v = rpm_version()
        ok( v )         -- v is defined
        ok( v ~= '' )   -- and not empty
    end )

end )

done()

end

-- end of file --
